# coding=utf-8
"""Utility functions for calculating PMV."""
from __future__ import division

from ladybug.rootfinding import secant
from ladybug.rootfinding import bisect

import math

FAILURE_MESSAGE = 'The following conditions caused a failure of PMV model ' \
    'convergence: Ta = {}; Tr = {}; Vel = {}; RH = {}; Met = {}; Clo= {}'


def predicted_mean_vote(ta, tr, vel, rh, met, clo, wme=0, still_air_threshold=0.1):
    """Calculate PMV using Fanger's original model and Pierce SET.

    This method uses the officially correct way to calculate PMV comfort according to
    the 2015 ASHRAE-55 Thermal Comfort Standard. This function will return
    correct values even if the air speed is above the sill air threshold of
    Fanger's original model (> 0.1 m/s).

    However, because this function always returns the Standard Effective Temperature
    (SET), it will run much more slowly for cases that are below the still air threshold
    (roughly 10 times as long). For cases where a SET output is not required, it
    is recommended that the predicted_mean_vote_no_set method be used.

    Note:
        [1] ASHRAE Standard 55 (2017). "Thermal Environmental Conditions
        for Human Occupancy".

        [2] Hoyt Tyler, Schiavon Stefano, Piccioli Alberto, Cheung Toby, Moon Dustin,
        and Steinfeld Kyle, 2017, CBE Thermal Comfort Tool. Center for the Built
        Environment, University of California Berkeley,
        http://comfort.cbe.berkeley.edu/

        [3] Doherty, T.J., and E.A. Arens.  (1988).  Evaluation of the physiological
        bases of thermal comfort models. ASHRAE Transactions, Vol. 94, Part 1, 15 pp.
        https://escholarship.org/uc/item/6pq3r5pr

    Args:
        ta: Air temperature [C]
        tr: Mean radiant temperature [C]
        vel: Relative air velocity [m/s]
        rh: Relative humidity [%]
        met: Metabolic rate [met]
        clo: Clothing [clo]
        wme: External work [met], normally around 0 when seated
        still_air_threshold: The air velocity in m/s at which the Pierce
            Standard Effective Temperature (SET) model will be used
            to correct values in the original Fanger PMV model.
            Default is 0.1 m/s per the 2015 release of ASHRAE Standard-55.

    Returns:
        A dictionary containing results of the PMV model with the following keys

        -   pmv -- Predicted mean vote (PMV)
        -   ppd -- Percent predicted dissatisfied (PPD) [%]
        -   set -- Standard effective temperature (SET) [C]
        -   ta_adj -- Air temperature adjusted for air speed [C]
        -   ce -- Cooling effect. The difference between the air temperature
            and the adjusted air temperature [C]
        -   heat_loss -- A dictionary with the 6 heat loss terms of the PMV model.
            The dictionary keys are as follows:

            -   cond -- heat loss by conduction [W]
            -   sweat -- heat loss by sweating [W]
            -   res_l -- heat loss by latent respiration [W]
            -   res_s -- heat loss by dry respiration [W]
            -   rad -- heat loss by radiation [W]
            -   conv -- heat loss by convection [W]
    """
    se_temp = pierce_set(ta, tr, vel, rh, met, clo, wme)

    if vel <= still_air_threshold:  # use the original Fanger model
        pmv, ppd, heat_loss = fanger_pmv(ta, tr, vel, rh, met, clo, wme)
        ta_adj, ce = ta, 0.
    else:  # use the SET model to correct the cooling effect in Fanger model
        ce_l = 0.
        ce_r = 40.
        eps = 0.001  # precision of ce

        def fn(ce):
            return se_temp - pierce_set(ta - ce, tr - ce, still_air_threshold,
                                        rh, met, clo, wme)

        try:
            ce = secant(ce_l, ce_r, fn, eps)
        except OverflowError:
            ce = None
        if ce is None:  # ce can be None because OverflowError or max secant iterations 
            ce = bisect(ce_l, ce_r, fn, eps, 0)

        pmv, ppd, heat_loss = fanger_pmv(
            ta - ce, tr - ce, still_air_threshold, rh, met, clo, wme)
        ta_adj = ta - ce

    result = {}
    result['pmv'] = pmv
    result['ppd'] = ppd
    result['set'] = se_temp
    result['ta_adj'] = ta_adj
    result['ce'] = ce
    result['heat_loss'] = heat_loss

    return result


def predicted_mean_vote_no_set(ta, tr, vel, rh, met, clo, wme=0, still_air_threshold=0.1):
    """Calculate PMV using Fanger's model and Pierce SET model ONLY WHEN NECESSARY.

    This method uses the officially correct way to calculate PMV comfort according to
    the 2015 ASHRAE-55 Thermal Comfort Standard. This function will return
    correct values even if the air speed is above the sill air threshold of
    Fanger's original equation (> 0.1 m/s).

    However, because this function does not return the Standard Effective Temperature
    (SET), it will run much faster for cases that are below the still air threshold
    (roughly 1/10th the time).

    Args:
        ta: Air temperature [C]
        tr: Mean radiant temperature [C]
        vel: Relative air velocity [m/s]
        rh: Relative humidity [%]
        met: Metabolic rate [met]
        clo: Clothing [clo]
        wme: External work [met], normally around 0 when seated
        still_air_threshold: The air velocity in m/s at which the Pierce
            Standard Effective Temperature (SET) model will be used
            to correct values in the original Fanger PMV model.
            Default is 0.1 m/s per the 2015 release of ASHRAE Standard-55.

    Returns:
        A dictionary containing results of the PMV model with the following keys

        -   pmv -- Predicted mean vote (PMV)
        -   ppd -- Percent predicted dissatisfied (PPD) [%]
        -   ta_adj -- Air temperature adjusted for air speed [C]
        -   ce -- Cooling effect. The difference between the air temperature
            and the adjusted air temperature [C]
        -   heat_loss -- A dictionary with the 6 heat loss terms of the PMV model.
            The dictionary keys are as follows:

            -   cond -- heat loss by conduction [W]
            -   sweat -- heat loss by sweating [W]
            -   res_l -- heat loss by latent respiration [W]
            -   res_s -- heat loss by dry respiration [W]
            -   rad -- heat loss by radiation [W]
            -   conv -- heat loss by convection [W]
    """
    if vel <= still_air_threshold:  # use the original Fanger model
        pmv, ppd, heat_loss = fanger_pmv(ta, tr, vel, rh, met, clo, wme)
        ta_adj, ce = ta, 0.
    else:  # use the SET model to correct the cooling effect in Fanger model
        ce_l = 0.
        ce_r = 40.
        eps = 0.001  # precision of ce
        se_temp = pierce_set(ta, tr, vel, rh, met, clo, wme)

        def fn(ce):
            return se_temp - pierce_set(ta - ce, tr - ce, still_air_threshold,
                                        rh, met, clo, wme)

        try:
            ce = secant(ce_l, ce_r, fn, eps)
        except OverflowError:
            ce = None
        if ce is None:  # ce can be None because OverflowError or max secant iterations 
            ce = bisect(ce_l, ce_r, fn, eps, 0)

        pmv, ppd, heat_loss = fanger_pmv(
            ta - ce, tr - ce, still_air_threshold, rh, met, clo, wme)
        ta_adj = ta - ce

    result = {}
    result['pmv'] = pmv
    result['ppd'] = ppd
    result['ta_adj'] = ta_adj
    result['ce'] = ce
    result['heat_loss'] = heat_loss

    return result


def fanger_pmv(ta, tr, vel, rh, met, clo, wme=0):
    """Calculate PMV using only Fanger's original equation.

    Note that Fanger's original experiments were conducted at
    low air speeds (<0.1 m/s) and the 2015 ASHRAE-55 thermal comfort
    standard states that one should use standard effective temperature
    (SET) to correct for the cooling effect of air speed in cases
    where such speeds exceed 0.1 m/s.  The pmv() function in this module
    will apply this SET correction in cases where it is appropriate.

    Note:
        [1] Fanger, P.O. (1970). Thermal Comfort: Analysis and applications
        in environmental engineering. Copenhagen: Danish Technical Press.

    Args:
        ta: Air temperature [C]
        tr: Mean radiant temperature [C]
        vel: Relative air velocity [m/s]
        rh: Relative humidity [%]
        met: Metabolic rate [met]
        clo: Clothing [clo]
        wme: External work [met], normally around 0 when seated

    Returns:
        A tuple with three elements

        -   pmv: Predicted mean vote (PMV)
        -   ppd: Percentage of people dissatisfied (PPD) [%]
        -   heat_loss: A dictionary with the 6 heat loss terms of the PMV model.
            The dictionary items are as follows:

            -   'cond': heat loss by conduction [W]
            -   'sweat': heat loss by sweating [W]
            -   'res_l': heat loss by latent respiration [W]
            -   'res_s' heat loss by dry respiration [W]
            -   'rad': heat loss by radiation [W]
            -   'conv' heat loss by convection [W]
    """

    pa = rh * 10. * math.exp(16.6536 - 4030.183 / (ta + 235.))

    icl = 0.155 * clo  # thermal insulation of the clothing in M2K/W
    m = met * 58.15  # metabolic rate in W/M2
    w = wme * 58.15  # external work in W/M2
    mw = m - w  # internal heat production in the human body
    if icl <= 0.078:
        fcl = 1 + (1.29 * icl)
    else:
        fcl = 1.05 + (0.645 * icl)

    # heat transfer coefficient by forced convection
    hcf = 12.1 * math.sqrt(vel)
    taa = ta + 273.
    tra = tr + 273.
    tcla = taa + (35.5 - ta) / (3.5 * icl + 0.1)

    p1 = icl * fcl
    p2 = p1 * 3.96
    p3 = p1 * 100.
    p4 = p1 * taa
    p5 = 308.7 - 0.028 * mw + (p2 * ((tra / 100.) ** 4))
    xn = tcla / 100.
    xf = tcla / 50.
    eps = 0.00015

    n = 0
    while abs(xn - xf) > eps:
        xf = (xf + xn) / 2.
        hcn = 2.38 * (abs(100.0 * xf - taa) ** 0.25)
        if hcf > hcn:
            hc = hcf
        else:
            hc = hcn
        xn = (p5 + p4 * hc - p2 * (xf ** 4)) / (100. + p3 * hc)
        n += 1
        if n > 150:
            print(FAILURE_MESSAGE.format(ta, tr, vel, rh, met, clo))
            return 0.0, 5.0, \
                {'cond': 0, 'sweat': 0, 'res_l': 0, 'res_s': 0, 'rad': 0, 'conv': 0}

    tcl = 100. * xn - 273.

    # heat loss conduction through skin
    hl1 = 3.05 * 0.001 * (5733. - (6.99 * mw) - pa)
    # heat loss by sweating
    if mw > 58.15:
        hl2 = 0.42 * (mw - 58.15)
    else:
        hl2 = 0
    # latent respiration heat loss
    hl3 = 1.7 * 0.00001 * m * (5867. - pa)
    # dry respiration heat loss
    hl4 = 0.0014 * m * (34. - ta)
    # heat loss by radiation
    hl5 = 3.96 * fcl * (math.pow(xn, 4) - math.pow(tra / 100., 4))
    # heat loss by convection
    hl6 = fcl * hc * (tcl - ta)

    ts = 0.303 * math.exp(-0.036 * m) + 0.028
    pmv = ts * (mw - hl1 - hl2 - hl3 - hl4 - hl5 - hl6)
    ppd = ppd_from_pmv(pmv)

    # collect heat loss terms.
    heat_loss = {
        'cond': hl1,
        'sweat': hl2,
        'res_l': hl3,
        'res_s': hl4,
        'rad': hl5,
        'conv': hl6
    }

    return pmv, ppd, heat_loss


def pierce_set(ta, tr, vel, rh, met, clo, wme=0.):
    """Calculate Standard Effective Temperature (SET).

    This function uses the J.B. Pierce two-node model of human thermoregulation.

    Note:
        [1] Nishi, Y; Gagge, A.P. (1977). "Effective temperature scale useful for
        hypo-and hyperbaric environments". Aviation, Space, and Environmental Medicine.
        48 (2): 97–107.

    Args:
        ta: Air temperature [C]
        tr: Mean radiant temperature [C]
        vel: Relative air velocity [m/s]
        rh: Relative humidity [%]
        met: Metabolic rate [met]
        clo: Clothing [clo]
        wme: External work [met], normally around 0 when seated

    Returns:
        se_temp -- Standard effective temperature [C]
    """

    # key initial variables
    vapor_pressure = (rh * saturated_vapor_pressure_torr(ta)) / 100.
    air_velocity = max(vel, 0.1)
    kclo = 0.25
    bodyweight = 69.9
    bodysurfacearea = 1.8258
    metfactor = 58.2
    sbc = 0.000000056697  # Stefan-Boltzmann constant (W/m2-K4)
    csw = 170.
    cdil = 120.
    cstr = 0.5

    temp_skin_neutral = 33.7  # setpoint (neutral) value for Tsk
    temp_core_neutral = 36.8  # setpoint value for Tcr
    temp_body_neutral = 36.49  # setpoint for Tb
    skin_blood_flow_neutral = 6.3  # neutral value for skin_blood_flow

    # INITIAL VALUES - start of 1st experiment
    temp_skin = temp_skin_neutral
    temp_core = temp_core_neutral
    skin_blood_flow = skin_blood_flow_neutral
    mshiv = 0.0
    alfa = 0.1
    esk = 0.1 * met

    # Start new experiment here (for graded experiments)
    # UNIT CONVERSIONS (from input variables)
    p = 101.325  # pressure of the atmosphere in kPa
    pressure_in_atmospheres = p * 0.009869
    ltime = 60
    rcl = 0.155 * clo

    facl = 1.0 + 0.15 * clo  # % INCreaSE IN BODY SURFACE Area DUE TO CLOTHING
    LR = 2.2 / pressure_in_atmospheres  # Lewis Relation is 2.2 at sea level
    RM = met * metfactor
    M = met * metfactor

    if clo <= 0:
        wcrit = 0.38 * pow(air_velocity, -0.29)
        icl = 1.0
    else:
        wcrit = 0.59 * pow(air_velocity, -0.08)
        icl = 0.45

    chc = 3.0 * pow(pressure_in_atmospheres, 0.53)
    chcV = 8.600001 * pow((air_velocity * pressure_in_atmospheres), 0.53)
    chc = max(chc, chcV)

    # initial estimate of Tcl
    chr = 4.7
    ctc = chr + chc
    ra = 1.0 / (facl * ctc)  # resistance of air layer to dry heat transfer
    top = (chr * tr + chc * ta) / ctc
    tcl = top + (temp_skin - top) / (ctc * (ra + rcl))

    # ========================  BEGIN ITERATION
    #
    # Tcl and chr are solved iteratively using: H(Tsk - To) = ctc(Tcl - To),
    # where H = 1 / (ra + Rcl) and ra = 1 / Facl * ctc

    tcl_old = False
    flag = True
    for i in range(ltime - 1):
        while abs(tcl - tcl_old) > 0.01:
            if flag:
                tcl_old = tcl
                chr = 4.0 * sbc * pow(((tcl + tr) / 2.0 + 273.15), 3.0) * 0.72
                ctc = chr + chc
                ra = 1.0 / (facl * ctc)  # resistance of air layer to dry heat transfer
                top = (chr * tr + chc * ta) / ctc
            tcl = (ra * temp_skin + rcl * top) / (ra + rcl)
            flag = True
        flag = False
        dry = (temp_skin - top) / (ra + rcl)
        hfcs = (temp_core - temp_skin) * (5.28 + 1.163 * skin_blood_flow)
        eres = 0.0023 * M * (44.0 - vapor_pressure)
        cres = 0.0014 * M * (34.0 - ta)
        scr = M - hfcs - eres - cres - wme
        ssk = hfcs - dry - esk
        tcsk = 0.97 * alfa * bodyweight
        tccr = 0.97 * (1 - alfa) * bodyweight
        dtsk = (ssk * bodysurfacearea) / (tcsk * 60.0)  # deg C per minute
        dtcr = scr * bodysurfacearea / (tccr * 60.0)  # deg C per minute
        temp_skin = temp_skin + dtsk
        temp_core = temp_core + dtcr
        TB = alfa * temp_skin + (1 - alfa) * temp_core
        sksig = temp_skin - temp_skin_neutral
        warms = (sksig > 0) * sksig
        colds = ((-1.0 * sksig) > 0) * (-1.0 * sksig)
        crsig = (temp_core - temp_core_neutral)
        warmc = (crsig > 0) * crsig
        coldc = ((-1.0 * crsig) > 0) * (-1.0 * crsig)
        bdsig = TB - temp_body_neutral
        warmb = (bdsig > 0) * bdsig
        skin_blood_flow = (skin_blood_flow_neutral + cdil *
                           warmc) / (1 + cstr * colds)
        if skin_blood_flow > 90.0:
            skin_blood_flow = 90.0
        if skin_blood_flow < 0.5:
            skin_blood_flow = 0.5
        regsw = csw * warmb * math.exp(warms / 10.7)
        if regsw > 500.0:
            regsw = 500.0
        ersw = 0.68 * regsw
        rea = 1.0 / (LR * facl * chc)  # evaporative resistance of air layer
        recl = rcl / (LR * icl)  # evaporative resistance of clothing (icl=.45)
        emax = (saturated_vapor_pressure_torr(
            temp_skin) - vapor_pressure) / (rea + recl)
        prsw = ersw / emax
        pwet = 0.06 + 0.94 * prsw
        edif = pwet * emax - ersw
        esk = ersw + edif
        if pwet > wcrit:
            pwet = wcrit
            prsw = wcrit / 0.94
            ersw = prsw * emax
            edif = 0.06 * (1.0 - prsw) * emax
            esk = ersw + edif
        if emax < 0:
            edif = 0
            ersw = 0
            pwet = wcrit
            prsw = wcrit
            esk = emax
        esk = ersw + edif
        mshiv = 19.4 * colds * coldc
        M = RM + mshiv
        alfa = 0.0417737 + 0.7451833 / (skin_blood_flow + .585417)

    # Define new heat flow terms, coefficients, and abbreviations
    hsk = dry + esk  # total heat loss from skin
    RN = M - wme  # net metabolic heat production
    ecomf = 0.42 * (RN - (1 * metfactor))
    if ecomf < 0.0:
        ecomf = 0.0  # from Fanger
    emax = emax * wcrit
    W = pwet
    pssk = saturated_vapor_pressure_torr(temp_skin)
    # Definition of ASHRAE standard environment... denoted "S"
    chrS = chr
    if met < 0.85:
        chcS = 3.0
    else:
        chcS = 5.66 * pow((met - 0.85), 0.39)
        if chcS < 3.0:
            chcS = 3.0

    ctcs = chcS + chrS
    rclos = 1.52 / ((met - wme / metfactor) + 0.6944) - 0.1835
    rcls = 0.155 * rclos
    facls = 1.0 + kclo * rclos
    fcls = 1.0 / (1.0 + 0.155 * facls * ctcs * rclos)
    ims = 0.45
    icls = ims * chcS / ctcs * (1 - fcls) / (chcS / ctcs - fcls * ims)
    ras = 1.0 / (facls * ctcs)
    reaS = 1.0 / (LR * facls * chcS)
    reclS = rcls / (LR * icls)
    hd_s = 1.0 / (ras + rcls)
    he_s = 1.0 / (reaS + reclS)

    # SET* (standardized humidity, clo, Pb, and chc)
    # determined using Newton's iterative solution
    # FNERRS is defined in the GENERAL SETUP section above

    delta = .0001
    dx = 100.0
    x_old = temp_skin - hsk / hd_s  # lower bound for SET
    while abs(dx) > .01:
        err1 = (hsk - hd_s * (temp_skin - x_old) - W * he_s *
                (pssk - 0.5 * saturated_vapor_pressure_torr(x_old)))
        err2 = (hsk - hd_s * (temp_skin - (x_old + delta)) - W * he_s *
                (pssk - 0.5 * saturated_vapor_pressure_torr((x_old + delta))))
        x = x_old - delta * err1 / (err2 - err1)
        dx = x - x_old
        x_old = x
    se_temp = x

    return se_temp


def saturated_vapor_pressure_torr(db_temp):
    """Calculate saturated vapor pressure (Torr) at temperature (C)

    This is used to synchronize the results of the Standard Effective temperature (SET)
    model with the results of the original Fanger model.
    """
    return math.exp(18.6686 - 4030.183 / (db_temp + 235.0))


def ppd_from_pmv(pmv):
    """Calculate the Percentage of People Dissatisfied (PPD) from PMV.

    Args:
        pmv: The predicted mean vote (PMV) for which you want to know the PPD.

    Returns:
        ppd -- The percentage of people dissatisfied (PPD) for the input PMV.
    """
    return 100.0 - 95.0 * math.exp(-0.03353 * pow(pmv, 4.0) - 0.2179 * pow(pmv, 2.0))


def pmv_from_ppd(ppd, pmv_up_bound=3, ppd_tolerance=0.001):
    """Calculate the two possible Predicted Mean Vote (PMV) values for a PPD value.

    Args:
        ppd: The percentage of people dissatisfied (PPD) for which you want to know
            the possible PMV.
        pmv_up_bound: The upper limit of PMV expected for the input PPD.  Putting
            in a good estimate here will help the model converge on a
            solution faster. Defaut = 3
        ppd_tolerance: The acceptable error in meeting the target PPD.  Default = 0.001.

    Returns:
        A tuple with two elements

        -   pmv_lower: The lower (cold) PMV value that will produce the input ppd.
        -   pmv_upper: The upper (hot) PMV value that will produce the input ppd.
    """
    assert ppd > 5 and ppd < 100, \
        'PPD value {}% is outside acceptable limits of the PMV model.'.format(ppd)

    def fn(pmv):
        return (100.0 - 95.0 * math.exp(
            -0.03353 * pow(pmv, 4.) - 0.2179 * pow(pmv, 2.0))) - ppd

    # Solve for the missing higher PMV value.
    pmv_upper = secant(0, pmv_up_bound, fn, ppd_tolerance)
    if pmv_upper is None:
        pmv_upper = bisect(0, pmv_up_bound, fn, ppd_tolerance, 0)
    pmv_lower = pmv_upper * -1

    return pmv_lower, pmv_upper


def ppd_threshold_from_comfort_class(comf_class):
    """Get acceptable PPD comfort threshold given the EN-15251 comfort class.

    Args:
        comf_class: An integer representing the EN-15251 comfort class.
            Choose from: 1, 2, 3 (the higher the class, the greater the PPD threshold)

    Returns:
        ppd_thresh -- Acceptable PPD threshold.
    """
    if comf_class == 1:
        ppd_thresh = 6
    elif comf_class == 2:
        ppd_thresh = 10
    elif comf_class == 3:
        ppd_thresh = 15
    else:
        raise ValueError('Comfort class {} is not an acceptable value. '
                         'Choose from: 1, 2, 3'.format(comf_class))
    return ppd_thresh


def calc_missing_pmv_input(target_pmv, pmv_inputs, low_bound=0., up_bound=100.,
                           tolerance=0.001, still_air_threshold=0.1):
    """Return the value of a missing_pmv_input given a target_pmv and the 6 other inputs.

    This is particularly useful when trying to draw comfort polygons on charts
    using the PMV model.

    Args:
        target_pmv: The target PMV that you are trying to produce from the inputs to
            the PMV model.
        pmv_inputs: A dictionary of 7 pmv inputs with the following keys:
            'ta', 'tr', 'vel', 'rh', 'met', 'clo', 'wme'.  Each key
            should correspond to a value that represents that pmv input
            but one of these inputs should have a value of None.
            The input corresponding to None will be solved for by this function.
            One can also input None for both 'ta' and 'tr' to solve for the operative
            temperature that meets the target_pmv. In this case, both 'ta' and 'tr'
            in the output dictionary will be the same.
            Example (solving for relative humidity):

            .. code-block:: python

                {
                'ta': 20,
                'tr': 20,
                'vel': 0.05,
                'rh': None,
                'met': 1.2,
                'clo': 0.75,
                'wme': 0
                }

        low_bound: The lowest possible value of the missing input you are tying to
            find. Putting in a good value here will help the model converge to a
            solution faster.
        up_bound: The highest possible value of the missing input you are tying to
            find. Putting in a good value here will help the model converge to a
            solution faster.
        tolerance: The acceptable error in the target_pmv. The default is set to 0.001
        still_air_threshold: The air velocity in m/s at which the Pierce
            Standard Effective Temperature (SET) model will be used
            to correct values in the original Fanger PMV model.
            Default is 0.1 m/s per the 2015 release of ASHRAE Standard-55.

    Returns:
        complete_pmv_inputs -- A copy of the pmv_inputs dictionary but with values
        for all inputs. The missing input to the PMV model will be filled by the value
        that returns the target_pmv.
    """
    assert len(pmv_inputs.keys()) == 7, \
        'pmv_inputs must have 7 keys. Got {}.'.format(len(pmv_inputs.keys()))

    # Determine the function that should be used given the missing input.
    if pmv_inputs['ta'] is None and pmv_inputs['tr'] is None:
        def fn(x):
            return predicted_mean_vote_no_set(
                x, x, pmv_inputs['vel'], pmv_inputs['rh'],
                pmv_inputs['met'], pmv_inputs['clo'], pmv_inputs['wme'],
                still_air_threshold)['pmv'] - target_pmv
        missing_key = ('ta', 'tr')
    elif pmv_inputs['ta'] is None:
        def fn(x):
            return predicted_mean_vote_no_set(
                x, pmv_inputs['tr'], pmv_inputs['vel'], pmv_inputs['rh'],
                pmv_inputs['met'], pmv_inputs['clo'], pmv_inputs['wme'],
                still_air_threshold)['pmv'] - target_pmv
        missing_key = 'ta'
    elif pmv_inputs['tr'] is None:
        def fn(x):
            return predicted_mean_vote_no_set(
                pmv_inputs['ta'], x, pmv_inputs['vel'], pmv_inputs['rh'],
                pmv_inputs['met'], pmv_inputs['clo'],  pmv_inputs['wme'],
                still_air_threshold)['pmv'] - target_pmv
        missing_key = 'tr'
    elif pmv_inputs['vel'] is None:
        def fn(x):
            return target_pmv - predicted_mean_vote_no_set(
                pmv_inputs['ta'], pmv_inputs['tr'], x, pmv_inputs['rh'],
                pmv_inputs['met'], pmv_inputs['clo'], pmv_inputs['wme'],
                still_air_threshold)['pmv']
        missing_key = 'vel'
    elif pmv_inputs['rh'] is None:
        def fn(x):
            return predicted_mean_vote_no_set(
                pmv_inputs['ta'], pmv_inputs['tr'], pmv_inputs['vel'], x,
                pmv_inputs['met'], pmv_inputs['clo'], pmv_inputs['wme'],
                still_air_threshold)['pmv'] - target_pmv
        missing_key = 'rh'
    elif pmv_inputs['met'] is None:
        def fn(x):
            return predicted_mean_vote_no_set(
                pmv_inputs['ta'], pmv_inputs['tr'], pmv_inputs['vel'],
                pmv_inputs['rh'], x, pmv_inputs['clo'], pmv_inputs['wme'],
                still_air_threshold)['pmv'] - target_pmv
        missing_key = 'met'
    elif pmv_inputs['clo'] is None:
        def fn(x):
            return predicted_mean_vote_no_set(
                pmv_inputs['ta'], pmv_inputs['tr'], pmv_inputs['vel'],
                pmv_inputs['rh'], pmv_inputs['met'], x, pmv_inputs['wme'],
                still_air_threshold)['pmv'] - target_pmv
        missing_key = 'clo'
    else:
        def fn(x):
            return predicted_mean_vote_no_set(
                pmv_inputs['ta'], pmv_inputs['tr'], pmv_inputs['vel'],
                pmv_inputs['rh'], pmv_inputs['met'], pmv_inputs['clo'], x,
                still_air_threshold)['pmv'] - target_pmv
        missing_key = 'wme'

    # Solve for the missing input using the function.
    missing_val = None
    if missing_key != 'clo':  # bisect is much better at finding reasonable clo values
        missing_val = secant(low_bound, up_bound, fn, tolerance)
    if missing_val is None:
        missing_val = bisect(low_bound, up_bound, fn, tolerance, 0)

    # copy and complete the input dictionary
    pmv_inputs = pmv_inputs.copy()
    if isinstance(missing_key, str):
        pmv_inputs[missing_key] = missing_val
    else:
        for key in missing_key:
            pmv_inputs[key] = missing_val
    return pmv_inputs



# coding=utf-8
"""Parameters for specifying acceptable thermal conditions using the Adaptive model."""
from __future__ import division
import re

from ._base import ComfortParameter
from ..adaptive import neutral_temperature_conditioned, \
    ashrae55_neutral_offset_from_ppd, en15251_neutral_offset_from_comfort_class


class AdaptiveParameter(ComfortParameter):
    """Parameters of Adaptive comfort.

    Args:
        ashrae55_or_en15251: A boolean to note whether to use the ASHRAE-55 neutral
            temperature function (True) or the EN-15251 function (False).
            Note that this input will also determine default values for many of
            the other properties of this object.
        neutral_offset:  The number of degrees Celsius from the neutral temperature
            where the input operative temperature is considered acceptable.
            The default is 2.5C when the neutral temperature function is ASHRAE-55
            and 3C when the neutral temperature function is EN-15251.
            You may want to use the set_neutral_offset_from_ppd() or the
            set_neutral_offset_from_comfort_class() methods on this object to set
            this value using ppd from the ASHRAE-55 standard or comfort classes
            from the EN-15251 standard respectively.
        avg_month_or_running_mean: A boolean to note whether the prevailing outdoor
            temperature is computed from the average monthly temperature (True) or
            a weighted running mean of the last week (False).  The default is True
            when the neutral temperature function is ASHRAE-55 and False when the
            neutral temperature function is EN-15251.
        discrete_or_continuous_air_speed: A boolean to note whether discrete
            categories should be used to assess the effect of elevated air speed
            (True) or whether a continuous function should be used (False).
            The default is True when the neutral temperature function is ASHRAE-55
            and False when the neutral temperature function is EN-15251.
        cold_prevail_temp_limit: A number indicating the prevailing outdoor
            temperature below which acceptable indoor operative temperatures
            flatline. The default is 10C when the neutral temperature function is
            ASHRAE-55 and 15C when the neutral temperature function is EN-15251.
            This number cannot be greater than 22C and cannot be less than 10C.
        conditioning: A number between 0 and 1 that represents how "conditioned" vs.
            "free-running" the building is.

            * 0 = free-running (completely passive with no air conditioning)
            * 1 = conditioned (no operable windows and fully air conditioned)

            The default is 0 since both the ASHRAE-55 and EN-15251 standards prohibit
            the use of adaptive comfort methods when a cooling system is active.

    Properties:
        * ashrae55_or_en15251
        * neutral_offset
        * avg_month_or_running_mean
        * discrete_or_continuous_air_speed
        * cold_prevail_temp_limit
        * conditioning
        * standard
        * prevailing_temperature_method
        * minimum_operative
    """
    _model = 'Adaptive'
    __slots__ = ('_standard', '_neutral_offset', '_prevail_method', '_air_speed_method',
                 '_cold_prevail_temp_limit', '_conditioning', '_min_operative')

    def __init__(self, ashrae55_or_en15251=None, neutral_offset=None,
                 avg_month_or_running_mean=None, discrete_or_continuous_air_speed=None,
                 cold_prevail_temp_limit=None, conditioning=None):
        """Initialize Adaptive Parameters.
        """
        # get the standard
        self._standard = ashrae55_or_en15251 if ashrae55_or_en15251 is not None else True
        assert isinstance(self._standard, bool), 'ashrae55_or_en15251 must be '\
            'a boolean. Got {}'.format(type(self._standard))

        # set defaults based on the standard
        if self._standard is True:
            default_neutral_offset = 2.5
            default_prevail_method = True
            default_air_speed_method = True
            default_cold_prevail_temp_limit = 10
        else:
            default_neutral_offset = 3
            default_prevail_method = False
            default_air_speed_method = False
            default_cold_prevail_temp_limit = 15

        # assign properties based on defaults and input
        self._prevail_method = avg_month_or_running_mean if \
            avg_month_or_running_mean is not None else default_prevail_method
        self._air_speed_method = discrete_or_continuous_air_speed if \
            discrete_or_continuous_air_speed is not None else default_air_speed_method
        self._cold_prevail_temp_limit = cold_prevail_temp_limit if \
            cold_prevail_temp_limit is not None else default_cold_prevail_temp_limit
        self._conditioning = conditioning if conditioning is not None else 0

        # perform range checks on the inputs
        assert 10 <= self._cold_prevail_temp_limit <= 22, \
            'cold_prevail_temp_limit must be between 10 and 22. Got {}'.format(
                self._cold_prevail_temp_limit)
        assert 0 <= self._conditioning <= 1, \
            'conditioning must be between 0 and 1. Got {}'.format(self._conditioning)

        # assign the neutral temperature offset
        self.neutral_offset = neutral_offset if \
            neutral_offset is not None else default_neutral_offset

    @classmethod
    def from_dict(cls, data):
        """Create a AdaptiveParameter object from a dictionary.

        Args:
            data: A AdaptiveParameter dictionary in following the format below.

        .. code-block:: python

            {
            'type': 'AdaptiveParameter',
            'ashrae55_or_en15251': True,
            'neutral_offset': 2.5,
            'avg_month_or_running_mean': False,
            'discrete_or_continuous_air_speed': True,
            'cold_prevail_temp_limit': 10,
            'conditioning': 0
            }
        """
        assert data['type'] == 'AdaptiveParameter', \
            'Expected AdaptiveParameter dictionary. Got {}.'.format(data['type'])
        ashrae55_or_en15251 = data['ashrae55_or_en15251'] if \
            'ashrae55_or_en15251' in data else None
        neutral_offset = data['neutral_offset'] if \
            'neutral_offset' in data else None
        avg_month_or_running_mean = data['avg_month_or_running_mean'] if \
            'avg_month_or_running_mean' in data else None
        discrete_or_continuous_air_speed = data['discrete_or_continuous_air_speed'] if \
            'discrete_or_continuous_air_speed' in data else None
        cold_prevail_temp_limit = data['cold_prevail_temp_limit'] if \
            'cold_prevail_temp_limit' in data else None
        conditioning = data['conditioning'] if \
            'conditioning' in data else None
        return cls(ashrae55_or_en15251, neutral_offset, avg_month_or_running_mean,
                   discrete_or_continuous_air_speed, cold_prevail_temp_limit,
                   conditioning)

    @classmethod
    def from_string(cls, adaptive_parameter_string):
        """Create an AdaptiveParameter object from an AdaptiveParameter string."""
        str_pattern = re.compile(r"\-\-(\S*\s\S*)")
        matches = str_pattern.findall(adaptive_parameter_string)
        par_dict = {item.split(' ')[0]: item.split(' ')[1] for item in matches}
        ashrae55 = True if 'standard' not in par_dict \
            or par_dict['standard'].upper() == 'ASHRAE-55' else False
        offset = float(par_dict['neutral-offset']) \
            if 'neutral-offset' in par_dict else None
        avg_month = None
        if 'prevail-method' in par_dict:
            avg_month = True if par_dict['prevail-method'].lower() == 'averagedmonthly' \
                else False
        spd_method = None
        if 'air-speed-method' in par_dict:
            spd_method = True if par_dict['air-speed-method'].lower() == 'discrete' \
                else False
        cold_limit = float(par_dict['cold-limit']) \
            if 'cold-limit' in par_dict else None
        conditioning = float(par_dict['conditioning']) \
            if 'conditioning' in par_dict else None
        return cls(ashrae55, offset, avg_month, spd_method, cold_limit, conditioning)

    @property
    def ashrae55_or_en15251(self):
        """A boolean to note whether to use the ASHRAE-55 neutral temperature
        function (True) or the EN-15251 function (False)."""
        return self._standard

    @property
    def neutral_offset(self):
        """The degrees Celsius from the neutral temperature where the operative
        temperature is considered acceptable."""
        return self._neutral_offset

    @neutral_offset.setter
    def neutral_offset(self, offset):
        assert 0 < offset <= 10, \
            'neutral_offset must be between 0 and 10 C. Got {}'.format(offset)
        self._neutral_offset = offset
        self._calc_min_operative_temperature()

    @property
    def avg_month_or_running_mean(self):
        """Boolean noting whether prevailing outdoor temperature is computed from the
        average monthly temperature (True) or a weighted running mean (False)."""
        return self._prevail_method

    @property
    def discrete_or_continuous_air_speed(self):
        """Boolean noting whether discrete categories are used to assess elevated
        air speed (True) or whether a continuous function is used (False)."""
        return self._air_speed_method

    @property
    def cold_prevail_temp_limit(self):
        """The prevailing outdoor temperature below which acceptable indoor
        operative temperatures flatline. [C]"""
        return self._cold_prevail_temp_limit

    @property
    def conditioning(self):
        """A decimal noting how conditioned(1) vs. free-running(0) the building is."""
        return self._conditioning

    @property
    def standard(self):
        """Text denoting the standard.

        Either 'ASHRAE-55' or 'EN-15251'
        """
        if self._standard is True:
            return 'ASHRAE-55'
        else:
            return 'EN-15251'

    @property
    def prevailing_temperature_method(self):
        """Text denoting previaling temperature method.

        Either 'Averaged Monthly' or 'Running Mean'.
        """
        if self._prevail_method is True:
            return 'AveragedMonthly'
        else:
            return 'RunningMean'

    @property
    def air_speed_method(self):
        """Text denoting the type of function used for the cooling effect of air speed.

        Either 'Discrete' or 'Continuous'.
        """
        if self._air_speed_method is True:
            return 'Discrete'
        else:
            return 'Continuous'

    @property
    def minimum_operative(self):
        """Operative Temperature in C below which conditions cannot be comfortable."""
        return self._min_operative

    def set_neutral_offset_from_ppd(self, ppd):
        """Set the offset from neutral temperature given the ASHRAE-55 PPD limit."""
        self.neutral_offset = ashrae55_neutral_offset_from_ppd(ppd)

    def set_neutral_offset_from_comfort_class(self, comfort_class):
        """Set the offset from neutral temperature given the EN-15251 comfort class."""
        self.neutral_offset = en15251_neutral_offset_from_comfort_class(comfort_class)

    def is_comfortable(self, comfort_result, cooling_effect=0):
        """Determine if conditions are comfortable or not.

        Values are one of the following:

        * 0 = uncomfortable
        * 1 = comfortable

        Args:
            comfort_result: An adaptive comfort result dictionary from the
                adaptive_comfort_ashrae55 or adaptive_comfort_en15251 functions.
            cooling_effect: Cooling effect from elevated air speed.
        """
        return 1 if (comfort_result['to'] >= self._min_operative and
                    comfort_result['deg_comf'] >= -self.neutral_offset and
                    comfort_result['deg_comf'] <= self.neutral_offset +
                    cooling_effect) else 0

    def thermal_condition(self, comfort_result, cooling_effect=0):
        """Determine whether conditions are cold, neutral or hot.

        Values are one of the following:

        * -1 = cold
        * 0 = netural
        * +1 = hot

        Args:
            comfort_result: An adaptive comfort result dictionary from the
                adaptive_comfort_ashrae55 or adaptive_comfort_en15251 functions.
            cooling_effect: Cooling effect from elevated air speed
        """
        if self.is_comfortable(comfort_result, cooling_effect) == 0:
            return 1 if comfort_result['deg_comf'] > 0 else -1
        else:
            return 0

    def to_dict(self):
        """AdaptiveParameter dictionary representation."""
        return {
            'type': 'AdaptiveParameter',
            'ashrae55_or_en15251': self.ashrae55_or_en15251,
            'neutral_offset': self.neutral_offset,
            'avg_month_or_running_mean': self.avg_month_or_running_mean,
            'discrete_or_continuous_air_speed': self.discrete_or_continuous_air_speed,
            'cold_prevail_temp_limit': self.cold_prevail_temp_limit,
            'conditioning': self.conditioning
        }

    def __copy__(self):
        return AdaptiveParameter(self.ashrae55_or_en15251, self.neutral_offset,
                                 self.avg_month_or_running_mean,
                                 self.discrete_or_continuous_air_speed,
                                 self.cold_prevail_temp_limit, self.conditioning)

    def _calc_min_operative_temperature(self):
        """Set operative temperature below which conditions cannot be comfortable."""
        self._min_operative = neutral_temperature_conditioned(
            self._cold_prevail_temp_limit, self._conditioning, self.standard) \
            - self._neutral_offset

    def __repr__(self):
        """Adaptive comfort parameters representation."""
        return '--standard {} --neutral-offset {} ' \
            '--prevail-method {} --air-speed-method {} ' \
            '--cold-limit {} --conditioning {}'.format(
                self.standard, self.neutral_offset, self.prevailing_temperature_method,
                self.air_speed_method, self.cold_prevail_temp_limit, self.conditioning)



# coding=utf-8
"""Utility functions for calculating the Wet Bulb Globe Temperature (WBGT)."""
from __future__ import division

import math

from ladybug.psychrometrics import saturated_vapor_pressure


def wet_bulb_globe_temperature(ta, mrt, ws, rh):
    """Get wet-bulb globe temperature (WBGT) for outdoor conditions.
    
    WBGT is a type of feels-like temperature that is widely used as a heat stress
    index (ISO 7243). It is incorporates the effect of temperature, humidity, wind
    speed, and mean radiant temperature (optionally including the effect of sun).

    The WBGT outdoor formula is based on a mathematical model published by
    K. Blazejczyk et al. in 2014 [1] with the natural wet bulb temperature
    (Tnwb) based on the code written by Nick Burns.

    Note:
        [1] Heat stress and occupational health and safety spatial
        and temporal differentiation, K. Blazejczyk, J.Baranowski, A. Blazejczyk,
        Miscellanea geographica regional studies on development, Vol. 18,
        No. 1, 2014

    Args:
        ta: Air temperature [C]
        mrt: Radiant temperature [C]
        ws: Wind speed at average human height [m/s]
        rh: Relative humidity [%]

    Returns:
        Outdoor WBGT
    """
    p_ws = saturated_vapor_pressure(ta + 273.15)  # saturation pressure
    vp = p_ws * (rh / 10000)  # partial pressure
    tnwb = -9.27522 + 0.70196 * ta + 0.30338 * vp + 0.07823 * rh  # natural wet bulb temp
    tg = 2.098 - 2.561 * ws + 0.5957 * ta + 0.4017 * mrt  # black-globe temperature
    wbgt_od = 0.7 * tnwb + 0.2 * tg + 0.1 * ta  # outdoor WBGT
    return wbgt_od


def wbgt_warning_category(wbgt):
    """Get the warning category associated with a given WBGT.

    Categories are based on the US National Weather Service (NWS), which issues the
    following suggested actions and impact prevention:

    * 0 = No Warning.
    * 1 = Working or exercising in direct sunlight will stress your body after
        45 minutes. Take at least 15 minutes of breaks each hour if working or
        exercising in direct sunlight.
    * 2 = Working or exercising in direct sunlight will stress your body after
        30 minutes. Take at least 30 minutes of breaks each hour if working or
        exercising in direct sunlight.
    * 3 = Working or exercising in direct sunlight will stress your body after
        20 minutes. Take at least 40 minutes of breaks each hour if working or
        exercising in direct sunlight.
    * 4 = Working or exercising in direct sunlight will stress your body after
        15 minutes. Take at least 45 minutes of breaks each hour if working or
        exercising in direct sunlight.

    Args:
        wbgt: Wet Bulb Globe Temperature [C].

    Returns:
        category: An integer indicating the level of warning associated with
        the Wet Bulb Globe Temperature (WBGT).
    """
    wbgt_f = wbgt * 9. / 5. + 32.  # convert to fahrenheit

    if wbgt_f < 80:
        category = 0
    elif wbgt_f < 85:
        category = 1
    elif wbgt_f < 88:
        category = 2
    elif wbgt_f < 90:
        category = 3
    else:
        category = 4

    return category
